# -*- coding: utf-8 -*-
from __future__ import unicode_literals

import logging
import os

from django.conf import settings
from django.core.files.storage import FileSystemStorage
from django.core.validators import MinLengthValidator
from django.db import models
from django.db.models.signals import post_delete
from django.dispatch import receiver
from django.utils.deconstruct import deconstructible

import django_drf_filepond.drf_filepond_settings as local_settings
from django.utils.functional import LazyObject
from django_drf_filepond.storage_utils import _get_storage_backend


LOG = logging.getLogger(__name__)

BASE_DIR = local_settings.BASE_DIR
try:
    from pathlib import Path
    if isinstance(BASE_DIR, Path):
        BASE_DIR = str(BASE_DIR)
except ImportError:
    pass

FILEPOND_UPLOAD_TMP = getattr(
    local_settings, 'UPLOAD_TMP',
    os.path.join(BASE_DIR, 'filepond_uploads'))

FILEPOND_FILE_STORE_PATH = getattr(local_settings, 'FILE_STORE_PATH', None)


@deconstructible
class FilePondUploadSystemStorage(FileSystemStorage):
    """
    Subclass FileSystemStorage to prevent creation of new migrations when
    using a file store location passed to FileSystemStorage using the
    location attribute. Instead the location is applied dynamically on
    creation of the subclass avoiding detection of changes by the migration
    system.

    Addresses #13. Fix is based on fix for similar issue in
    https://github.com/julen/pootle/commit/fd7800050172549e9f31544843b986691290ddc2
    """

    def __init__(self, **kwargs):
        kwargs.update({
            'location': FILEPOND_UPLOAD_TMP,
        })
        super(FilePondUploadSystemStorage, self).__init__(**kwargs)


storage = FilePondUploadSystemStorage()


@deconstructible
class FilePondLocalStoredStorage(FileSystemStorage):
    """
    Subclass FileSystemStorage to prevent creation of new migrations when
    using a file store location passed to FileSystemStorage using the
    location attribute. Instead the location is applied dynamically on
    creation of the subclass avoiding detection of changes by the migration
    system.

    Addresses #13. Fix is based on fix for similar issue in
    https://github.com/julen/pootle/commit/fd7800050172549e9f31544843b986691290ddc2
    """

    def __init__(self, **kwargs):
        kwargs.update({
            'location': FILEPOND_FILE_STORE_PATH,
        })
        super(FilePondLocalStoredStorage, self).__init__(**kwargs)


class DrfFilePondStoredStorage(LazyObject):

    def _setup(self):
        # Work out which storage backend we need to use and then
        # instantiate it and assign it to self._wrapped
        storage_module_name = getattr(local_settings, 'STORAGES_BACKEND', None)
        LOG.debug('Initialising storage backend with storage module name [%s]'
                  % storage_module_name)
        storage_backend = _get_storage_backend(storage_module_name)
        if not storage_backend:
            self._wrapped = FilePondLocalStoredStorage()
        else:
            self._wrapped = storage_backend


def get_upload_path(instance, filename):
    return os.path.join(instance.upload_id, filename)


class TemporaryUpload(models.Model):

    FILE_DATA = 'F'
    URL = 'U'
    UPLOAD_TYPE_CHOICES = (
        (FILE_DATA, 'Uploaded file data'),
        (URL, 'Remote file URL'),
    )

    # The unique ID returned to the client and the name of the temporary
    # directory created to hold file data
    upload_id = models.CharField(primary_key=True, max_length=22,
                                 validators=[MinLengthValidator(22)])
    # The unique ID used to store the file itself
    file_id = models.CharField(max_length=22,
                               validators=[MinLengthValidator(22)])
    file = models.FileField(storage=storage, upload_to=get_upload_path)
    upload_name = models.CharField(max_length=512)
    uploaded = models.DateTimeField(auto_now_add=True)
    upload_type = models.CharField(max_length=1,
                                   choices=UPLOAD_TYPE_CHOICES)
    uploaded_by = models.ForeignKey(settings.AUTH_USER_MODEL, null=True,
                                    blank=True, on_delete=models.CASCADE)

    def get_file_path(self):
        return self.file.path


class TemporaryUploadChunked(models.Model):
    # The unique ID returned to the client and the name of the temporary
    # directory created to hold file data - this will be re-used in the
    # main TemporaryUpload record for this upload if/when all the chunks have
    # been successfully received and the upload is finalised.
    upload_id = models.CharField(primary_key=True, max_length=22,
                                 validators=[MinLengthValidator(22)])
    # The unique ID used to store the file chunks which are appended with
    # _<chunk_num>
    file_id = models.CharField(max_length=22,
                               validators=[MinLengthValidator(22)])
    upload_dir = models.CharField(max_length=512, default=upload_id)
    last_chunk = models.IntegerField(default=0)
    offset = models.BigIntegerField(default=0)
    total_size = models.BigIntegerField(default=0)
    upload_name = models.CharField(max_length=512, default='')
    upload_complete = models.BooleanField(default=False)
    last_upload_time = models.DateTimeField(auto_now=True)
    uploaded_by = models.ForeignKey(settings.AUTH_USER_MODEL, null=True,
                                    blank=True, on_delete=models.CASCADE)


class StoredUpload(models.Model):

    # The unique upload ID assigned to this file when it was originally
    # uploaded (or retrieved from a remote URL)
    upload_id = models.CharField(primary_key=True, max_length=22,
                                 validators=[MinLengthValidator(22)])
    # The file name and path (relative to the base file store directory
    # as set by DJANGO_DRF_FILEPOND_FILE_STORE_PATH).
    file = models.FileField(storage=DrfFilePondStoredStorage(),
                            max_length=2048)
    uploaded = models.DateTimeField()
    stored = models.DateTimeField(auto_now_add=True)
    uploaded_by = models.ForeignKey(settings.AUTH_USER_MODEL, null=True,
                                    blank=True, on_delete=models.CASCADE)

    def get_absolute_file_path(self):
        fsp = local_settings.FILE_STORE_PATH
        if not fsp:
            fsp = ''
        return os.path.join(fsp, self.file.name)


# When a TemporaryUpload record is deleted, we need to delete the
# corresponding file from the filesystem by catching the post_delete signal.
@receiver(post_delete, sender=TemporaryUpload)
def delete_temp_upload_file(sender, instance, **kwargs):
    # Check that the file parameter for the instance is not None
    # and that the file exists and is not a directory! Then we can delete it
    LOG.debug('*** post_delete signal handler called. Deleting file.')
    if instance.file:
        if (os.path.exists(instance.file.path) and
                os.path.isfile(instance.file.path)):
            os.remove(instance.file.path)

    if local_settings.DELETE_UPLOAD_TMP_DIRS:
        file_dir = os.path.join(storage.location, instance.upload_id)
        if(os.path.exists(file_dir) and os.path.isdir(file_dir)):
            os.rmdir(file_dir)
            LOG.debug('*** post_delete signal handler called. Deleting temp '
                      'dir that contained file.')



import logging
import os

from django.core.files.uploadedfile import UploadedFile, InMemoryUploadedFile
from rest_framework import status
from rest_framework.exceptions import ParseError, MethodNotAllowed
from rest_framework.response import Response

from django_drf_filepond.models import TemporaryUpload, storage,\
    TemporaryUploadChunked
from io import BytesIO, StringIO
from django_drf_filepond.utils import _get_user
from six import text_type, binary_type

# There's no built in FileNotFoundError in Python 2
try:
    FileNotFoundError
except NameError:
    FileNotFoundError = IOError

LOG = logging.getLogger(__name__)


class FilepondFileUploader(object):

    @classmethod
    def get_uploader(cls, request):
        # Process the request to identify if it's a standard upload request
        # or a request that is related to a chunked upload. Return the right
        # kind of uploader to handle this.
        if request.method == 'PATCH':
            return FilepondChunkedFileUploader()
        if request.method == 'HEAD':
            return FilepondChunkedFileUploader()
        elif request.method == 'POST':
            file_obj = cls._get_file_obj(request)
            if (file_obj == '{}' and
                    request.META.get('HTTP_UPLOAD_LENGTH', None)):

                LOG.debug('Returning CHUNKED uploader to handle '
                          'upload request... ')
                return FilepondChunkedFileUploader()
        else:
            raise MethodNotAllowed('%s is an invalid method type'
                                   % (request.method))

        # If we didn't identify the need for a chunked uploader in any of the
        # above tests, treat this as a standard upload
        LOG.debug('Returning STANDARD uploader to handle upload request... ')
        return FilepondStandardFileUploader()

    @classmethod
    def _get_file_obj(cls, request):
        # By default the upload element name is expected to be "filepond"
        # As raised in issue #4, there are cases where there may be more
        # than one filepond instance on a page, or the developer has opted
        # not to use the name "filepond" for the filepond instance.
        # Using the example from #4, this provides support these cases.
        upload_field_name = 'filepond'
        if 'fp_upload_field' in request.data:
            upload_field_name = request.data['fp_upload_field']

        if upload_field_name not in request.data:
            raise ParseError('Invalid request data has been provided.')

        # The content of the upload field is a django.http.QueryDict.
        # The dict may have multiple values for a given field name.
        # When accessing the QueryDict by field name, e.g.
        # request.data['filepond'], if there is more than one value provided
        # for the requested field name, the last one is returned, to get all
        # values we use getlist(). For the file upload being handled here, we
        # expect the file field to contain two values, the first is the file
        # metadata, the second is the file object.
        upload_fields = request.data.getlist(upload_field_name)
        if len(upload_fields) == 1:
            file_obj = upload_fields[0]
        elif len(upload_fields) == 2:
            # file_metadata = upload_fields[0]
            file_obj = upload_fields[1]
        else:
            raise ParseError('Invalid number of fields in request data.')

        return file_obj

    # The file ID and upload ID are generated by _get_file_id in
    # django_drf_filepond.views. The TemporaryUpload model should validate
    # that the values provided are within spec but in some cases, e.g. when
    # using SQLite, that doesn't happen. We therefore provide these two
    # methods for doing local validation of these values since they are
    # passed in as parameters to handle_upload.
    @classmethod
    def _file_id_valid(cls, file_id):
        if isinstance(file_id, text_type) and (len(file_id) == 22):
            return True
        return False

    @classmethod
    def _upload_id_valid(cls, upload_id):
        if isinstance(upload_id, text_type) and (len(upload_id) == 22):
            return True
        return False


class FilepondStandardFileUploader(FilepondFileUploader):

    def handle_upload(self, request, upload_id, file_id):
        # Since the upload_id and file_id are being provided here as
        # parameters, we check that they are valid. This should be done by
        # the DB and an error would be generated in the tu.save() call below
        # however SQLite doesn't handle field length validation so this won't
        # be picked up when using SQLite.
        if ((not self._file_id_valid(file_id)) or
                (not self._upload_id_valid(upload_id))):
            return Response('Invalid ID for handling upload.',
                            content_type='text/plain',
                            status=status.HTTP_500_INTERNAL_SERVER_ERROR)

        file_obj = self._get_file_obj(request)

        # The type of parsed data should be a descendant of an UploadedFile
        # type.
        if not isinstance(file_obj, UploadedFile):
            raise ParseError('Invalid data type has been parsed.')

        # Save original file name and set name of saved file to the unique ID
        upload_filename = file_obj.name
        file_obj.name = file_id

        # Before we attempt to save the file, make sure that the upload
        # directory we're going to save to exists.
        # *** It's not necessary to explicitly create the directory since
        # *** the FileSystemStorage object creates the directory on save
        # if not os.path.exists(storage.location):
        #    LOG.debug('Filepond app: Creating file upload directory '
        #             '<%s>...' % storage.location)
        #    os.makedirs(storage.location, mode=0o700)

        LOG.debug('About to store uploaded temp file with filename: %s'
                  % (upload_filename))

        # We now need to create the temporary upload object and store the
        # file and metadata.
        tu = TemporaryUpload(upload_id=upload_id, file_id=file_id,
                             file=file_obj, upload_name=upload_filename,
                             upload_type=TemporaryUpload.FILE_DATA,
                             uploaded_by=_get_user(request))
        tu.save()

        response = Response(upload_id, status=status.HTTP_200_OK,
                            content_type='text/plain')

        return response


# Handles chunked file uploads as per the approach described in filepond's
# docs at https://pqina.nl/filepond/docs/patterns/api/server/#process-chunks
class FilepondChunkedFileUploader(FilepondFileUploader):

    def handle_upload(self, request, upload_id, file_id=None):
        # Since the upload_id is being provided here as a paramter, we check
        # it is valid. This should be done by the DB but for some DBs, e.g.
        # SQLite field length validation isn't handled. The same check is
        # done for file_id in the case of POST requests.
        if not self._upload_id_valid(upload_id):
            return Response('Invalid ID for handling upload.',
                            status=status.HTTP_500_INTERNAL_SERVER_ERROR)

        if request.method == 'PATCH':
            return self._handle_chunk_upload(request, upload_id)
        elif request.method == 'HEAD':
            return self._handle_chunk_restart(request, upload_id)
        elif request.method == 'POST':
            if not self._file_id_valid(file_id):
                return Response('Invalid ID for handling upload.',
                                status=status.HTTP_500_INTERNAL_SERVER_ERROR)
            return self._handle_new_chunk_upload(request, upload_id, file_id)

    def _handle_new_chunk_upload(self, request, upload_id, file_id):
        LOG.debug('Processing a new chunked upload request...')
        file_obj = self._get_file_obj(request)
        if file_obj != '{}':
            return Response('An invalid file object has been received '
                            'for a new chunked upload request.',
                            status=status.HTTP_400_BAD_REQUEST)

        ulen = request.META.get('HTTP_UPLOAD_LENGTH', None)
        if not ulen:
            return Response('No length for new chunked upload request.',
                            status=status.HTTP_400_BAD_REQUEST)

        LOG.debug('Handling a new chunked upload request for an upload '
                  'with total length %s bytes' % (ulen))

        # Do some general checks to make sure that the storage location
        # exists and that we're not being made to try and store something
        # outside the base storage location. Then create the new
        # temporary directory into which chunks will be stored
        base_loc = storage.base_location
        chunk_dir = os.path.abspath(os.path.join(base_loc, upload_id))
        if not chunk_dir.startswith(base_loc):
            return Response('Unable to create storage for upload data.',
                            status=status.HTTP_500_INTERNAL_SERVER_ERROR)
        if os.path.exists(base_loc):
            try:
                os.makedirs(chunk_dir)
            except OSError as e:
                LOG.debug('Unable to create chunk storage dir: %s' %
                          (str(e)))
                return Response(
                    'Unable to prepare storage for upload data.',
                    status=status.HTTP_500_INTERNAL_SERVER_ERROR)
        else:
            LOG.debug('The base data store location <%s> doesn\'t exist.'
                      ' Unable to create chunk dir.' % (base_loc))
            return Response('Data storage error occurred.',
                            status=status.HTTP_500_INTERNAL_SERVER_ERROR)

        # We now create the temporary chunked upload object
        # this will be updated as we receive the chunks.
        tuc = TemporaryUploadChunked(upload_id=upload_id, file_id=file_id,
                                     upload_dir=upload_id, total_size=ulen,
                                     uploaded_by=_get_user(request))
        tuc.save()

        return Response(upload_id, status=status.HTTP_200_OK,
                        content_type='text/plain')

    def _handle_chunk_upload(self, request, chunk_id):
        # Check that the incoming data can be accessed. If the request
        # content type was invalid then we want to raise an error here
        # Trying to access request data should result in a 415 response if
        # the data couldn't be handled by the configured parser.
        file_data = request.data

        if (not chunk_id) or (chunk_id == ''):
            return Response('A required chunk parameter is missing.',
                            status=status.HTTP_400_BAD_REQUEST)

        if isinstance(file_data, binary_type):
            fd = BytesIO(file_data)
        elif isinstance(file_data, text_type):
            fd = StringIO(file_data)
        else:
            return Response('Upload data type not recognised.',
                            status=status.HTTP_400_BAD_REQUEST)

        # Try to load a temporary chunked upload object for the provided id
        try:
            tuc = TemporaryUploadChunked.objects.get(upload_id=chunk_id)
        except TemporaryUploadChunked.DoesNotExist:
            return Response('Invalid chunk upload request data',
                            status=status.HTTP_400_BAD_REQUEST)

        # Get the required header information to handle the new data
        uoffset = request.META.get('HTTP_UPLOAD_OFFSET', None)
        ulength = request.META.get('HTTP_UPLOAD_LENGTH', None)
        uname = request.META.get('HTTP_UPLOAD_NAME', None)

        if (uoffset is None) or (ulength is None) or (uname is None):
            return Response('Chunk upload is missing required metadata',
                            status=status.HTTP_400_BAD_REQUEST)
        if int(ulength) != tuc.total_size:
            return Response('ERROR: Upload metadata is invalid - size changed',
                            status=status.HTTP_400_BAD_REQUEST)

        # if this is the first chunk, store the filename
        if tuc.last_chunk == 0:
            tuc.upload_name = uname
        else:
            if tuc.upload_name != uname:
                return Response('Chunk upload file metadata is invalid',
                                status=status.HTTP_400_BAD_REQUEST)

        LOG.debug('Handling chunk <%s> for upload id <%s> with name <%s> '
                  'size <%s> and offset <%s>...'
                  % (tuc.last_chunk+1, chunk_id, uname, ulength, uoffset))

        LOG.debug('Current length and offset in the record is: length <%s> '
                  '  offset <%s>' % (tuc.total_size, tuc.offset))

        # Check that our recorded offset matches the offset provided by the
        # client...if not, there's an error.
        if not (int(uoffset) == tuc.offset):
            LOG.error('Offset provided by client <%s> doesn\'t match the '
                      'stored offset <%s> for chunked upload id <%s>'
                      % (uoffset, tuc.offset, chunk_id))
            return Response('ERROR: Chunked upload metadata is invalid.',
                            status=status.HTTP_400_BAD_REQUEST)

        file_data_len = len(file_data)
        LOG.debug('Got data from request with length %s bytes'
                  % (file_data_len))

        # Store the chunk and check if we've now completed the upload
        upload_dir = os.path.join(storage.base_location, tuc.upload_dir)
        upload_file = os.path.join(tuc.upload_dir,
                                   '%s_%s' % (tuc.file_id, tuc.last_chunk+1))
        if not os.path.exists(upload_dir):
            return Response('Chunk storage location error',
                            status=status.HTTP_500_INTERNAL_SERVER_ERROR)

        storage.save(upload_file, fd)
        # Set the updated chunk number and the new offset
        tuc.last_chunk = tuc.last_chunk + 1
        tuc.offset = tuc.offset + file_data_len
        if tuc.offset == tuc.total_size:
            tuc.upload_complete = True
        tuc.save()

        # At this point, if the upload is complete, we can rebuild the chunks
        # into the complete file and store it with a TemporaryUpload object.
        if tuc.upload_complete:
            try:
                self._store_upload(tuc)
            except (ValueError, FileNotFoundError) as e:
                LOG.error('Error storing upload: %s' % (str(e)))
                return Response('Error storing uploaded file.',
                                status=status.HTTP_500_INTERNAL_SERVER_ERROR)

        return Response(chunk_id, status=status.HTTP_200_OK,
                        content_type='text/plain')

    def _store_upload(self, tuc):
        if not tuc.upload_complete:
            LOG.error('Attempt to store an incomplete upload with ID <%s>'
                      % (tuc.upload_id))
            raise ValueError('Attempt to store an incomplete upload with ID '
                             '<%s>' % (tuc.upload_id))

        # Load each of the file parts into a BytesIO object and store them
        # via a TemporaryUpload object.
        chunk_dir = os.path.join(storage.base_location, tuc.upload_dir)
        file_data = BytesIO()
        for i in range(1, tuc.last_chunk+1):
            chunk_file = os.path.join(chunk_dir, '%s_%s' % (tuc.file_id, i))
            if not os.path.exists(chunk_file):
                raise FileNotFoundError('Chunk file not found for chunk <%s>'
                                        % (i))

            with open(chunk_file, 'rb') as cf:
                file_data.write(cf.read())

        # Prepare an InMemoryUploadedFile object so that the data can be
        # successfully saved via the FileField in the TemporaryUpload object
        memfile = InMemoryUploadedFile(file_data, None, tuc.file_id,
                                       'application/octet-stream',
                                       tuc.total_size, None)
        tu = TemporaryUpload(upload_id=tuc.upload_id, file_id=tuc.file_id,
                             file=memfile, upload_name=tuc.upload_name,
                             upload_type=TemporaryUpload.FILE_DATA,
                             uploaded_by=tuc.uploaded_by)
        tu.save()

        # Check that the final file is stored and of the correct size
        stored_file_path = os.path.join(chunk_dir, tuc.file_id)
        if ((not os.path.exists(stored_file_path)) or
                (not os.path.getsize(stored_file_path) == tuc.total_size)):
            raise ValueError('Stored file size wrong or file not found.')

        LOG.debug('Full file built from chunks and saved. Deleting chunks '
                  'and TemporaryUploadChunked object.')

        for i in range(1, tuc.last_chunk+1):
            chunk_file = os.path.join(chunk_dir, '%s_%s' % (tuc.file_id, i))
            os.remove(chunk_file)
        tuc.delete()

    def _handle_chunk_restart(self, request, upload_id):
        try:
            tuc = TemporaryUploadChunked.objects.get(upload_id=upload_id)
        except TemporaryUploadChunked.DoesNotExist:
            return Response('Invalid upload ID specified.',
                            status=status.HTTP_404_NOT_FOUND,
                            content_type='text/plain')

        if tuc.upload_complete is True:
            return Response('Invalid upload ID specified.',
                            status=status.HTTP_400_BAD_REQUEST,
                            content_type='text/plain')

        # Check that the directory for the chunks exists
        if not os.path.exists(os.path.join(storage.base_location,
                                           tuc.upload_dir)):
            return Response('Invalid upload location, can\'t continue upload.',
                            status=status.HTTP_500_INTERNAL_SERVER_ERROR,
                            content_type='text/plain')

        # TODO: Is it necessary to check for the existence of all previous
        #       chunk files here?
        LOG.debug('Returning offset to continue chunked upload. We have <%s> '
                  'chunks so far and are at offest <%s>.'
                  % (tuc.last_chunk, tuc.offset))
        return Response(upload_id, status=status.HTTP_200_OK,
                        headers={'Upload-Offset': str(tuc.offset)},
                        content_type='text/plain')



# -*- coding: utf-8 -*-
from __future__ import unicode_literals

import importlib
import logging
import mimetypes

import django_drf_filepond.drf_filepond_settings as local_settings
import os
import re
import requests
import shortuuid
import django_drf_filepond
from django.core.exceptions import ValidationError
from django.core.files.uploadedfile import InMemoryUploadedFile
from django.core.validators import URLValidator
from django.http.response import HttpResponse, HttpResponseNotFound, \
    HttpResponseServerError
from django_drf_filepond.api import get_stored_upload, \
    get_stored_upload_file_data
from django_drf_filepond.exceptions import ConfigurationError
from django_drf_filepond.models import TemporaryUpload, storage, StoredUpload
from django_drf_filepond.parsers import PlainTextParser, UploadChunkParser
from django_drf_filepond.renderers import PlainTextRenderer
from io import BytesIO
from requests.exceptions import ConnectionError
from rest_framework import status
from rest_framework.exceptions import ParseError, NotFound
from rest_framework.parsers import MultiPartParser
from rest_framework.response import Response
from rest_framework.views import APIView
from django_drf_filepond.uploaders import FilepondFileUploader
from django_drf_filepond.utils import _get_file_id, _get_user,\
    get_local_settings_base_dir

LOG = logging.getLogger(__name__)

LOAD_RESTORE_PARAM_NAME = 'id'


# There's no built in FileNotFoundError in Python 2
try:
    FileNotFoundError
except NameError:
    FileNotFoundError = IOError


# FIXME: This is a very basic approach to working out the MIME type.
#        It is prone to errors and can be inaccurate since it is
#        based only on the file extension.
#        A better approach would be to use python-magic but this
#        introduces another dependency and will likely result in
#        issues on some platforms.
#        Another option is to store the MIME type in the DB when the
#        file is uploaded and look this up and return it.
#
# At present this helper function takes only the filename (the data
# parameter). If this is refactored to use something like python-magic,
# 'data' will be the header data from the file to enable file type detection.
def _get_content_type(data, temporary=True):
    return mimetypes.guess_type(data)[0]


def _import_permission_classes(endpoint):
    """
    Iterates over array of string representations of permission classes from
    settings specified.
    """
    permission_classes = []
    if endpoint in local_settings.PERMISSION_CLASSES.keys():
        for perm_str in local_settings.PERMISSION_CLASSES[endpoint]:
            (modname, clname) = perm_str.rsplit('.', 1)
            mod = importlib.import_module(modname)
            class_ = getattr(mod, clname)
            permission_classes.append(class_)
    return permission_classes


class ProcessView(APIView):
    '''
    This view receives an uploaded file from the filepond client. It
    stores the file in a temporary location and generates a unique ID which
    it associates with the temporary upload. The unique ID is returned to
    the client. If and when the parent form is finally submitted, the unique
    ID is provided and the file is then moved from the temporary store into
    permanent storage in line with the requirements of the parent application.
    '''
    # This view uses the MultiPartParser to parse the uploaded file data
    # from FilePond.
    parser_classes = (MultiPartParser,)
    renderer_classes = (PlainTextRenderer,)
    permission_classes = _import_permission_classes('POST_PROCESS')

    def post(self, request):
        LOG.debug('Filepond API: Process view POST called...')

        # Check that the temporary upload directory has been set
        if not hasattr(local_settings, 'UPLOAD_TMP'):
            return Response('The file upload path settings are not '
                            'configured correctly.',
                            status=status.HTTP_500_INTERNAL_SERVER_ERROR)

        # By default, enforce that the temporary upload location must be a
        # sub-directory of the project base directory.
        # TODO: Check whether this is necessary - maybe add a security
        # parameter that can be disabled to turn off this check if the
        # developer wishes?
        LOCAL_BASE_DIR = get_local_settings_base_dir()
        if ((not (storage.location).startswith(LOCAL_BASE_DIR)) and
                (LOCAL_BASE_DIR !=
                 os.path.dirname(django_drf_filepond.__file__))):
            if not local_settings.ALLOW_EXTERNAL_UPLOAD_DIR:
                return Response('The file upload path settings are not '
                                'configured correctly.',
                                status=status.HTTP_500_INTERNAL_SERVER_ERROR)

        # Check that a relative path is not being used to store the
        # upload outside the specified UPLOAD_TMP directory.
        if not getattr(local_settings, 'UPLOAD_TMP').startswith(
                os.path.abspath(storage.location)):
            return Response('An invalid storage location has been '
                            'specified.',
                            status=status.HTTP_500_INTERNAL_SERVER_ERROR)

        # Check that we've received a file and then generate a unique ID
        # for it. Also generate a unique UD for the temp upload dir
        file_id = _get_file_id()
        upload_id = _get_file_id()

        try:
            uploader = FilepondFileUploader.get_uploader(request)
            response = uploader.handle_upload(request, file_id, upload_id)
        except ParseError as e:
            # Re-raise the ParseError to trigger a 400 response via DRF.
            raise e

        return response


class PatchView(APIView):
    '''
    This view handles a PATCH request containing a file chunk as part of the
    filepond chunked upload support. The chunk will relate to an existing
    chunked upload configuration created when a new chunked upload request
    was made to the ProcessView.
    See: https://pqina.nl/filepond/docs/patterns/api/server/#process-chunks
    for details of how chunked uploads are handled in filepond.
    Assuming everything in the request is valid, this view will store the
    chunk.
    '''
    # Chunk upload PATCH requests use the application/offset+octet-stream
    # Content-Type. Since this is different to the multipart upload used for
    # process requests, this is being handled in a separate view.
    # from FilePond.
    parser_classes = (UploadChunkParser,)
    renderer_classes = (PlainTextRenderer,)
    permission_classes = _import_permission_classes('PATCH_PATCH')

    def patch(self, request, chunk_id):
        LOG.debug('Filepond API: Patch view PATCH called...')
        uploader = FilepondFileUploader.get_uploader(request)
        return uploader.handle_upload(request, chunk_id)

    # The HEAD method is used to request information about a partially
    # completed chunked upload. Assuming the details are found and returned
    # correctly, the client will use this to restart a failed chunked upload.
    def head(self, request, chunk_id):
        LOG.debug('Filepond API: Patch view HEAD called...')
        uploader = FilepondFileUploader.get_uploader(request)
        return uploader.handle_upload(request, chunk_id)


class RevertView(APIView):

    parser_classes = (PlainTextParser,)
    renderer_classes = (PlainTextRenderer,)
    permission_classes = _import_permission_classes('DELETE_REVERT')
    '''
    This is called when we need to revert the uploaded file - i.e. undo is
    pressed and we remove the previously uploaded temporary file.
    '''
    def delete(self, request):
        # If we've received the incoming data as bytes, we need to decode
        # it to a string
        if isinstance(request.data, bytes):
            request_data = request.data.decode('utf-8')
        else:
            request_data = request.data

        # Expecting a 22-character unique ID telling us which temporary
        # upload to remove.
        LOG.debug('Filepond API: Revert view DELETE called...')
        upload_id = request_data.strip()

        if len(upload_id) != 22:
            raise ParseError('The provided data is invalid.')

        # Lookup the temporary file record
        try:
            tu = TemporaryUpload.objects.get(upload_id=upload_id)
            LOG.debug('About to delete temporary upload <%s> with original '
                      'filename <%s>' % (tu.upload_id, tu.upload_name))
            tu.delete()
        except TemporaryUpload.DoesNotExist:
            raise NotFound('The specified file does not exist.')

        return Response(status=status.HTTP_204_NO_CONTENT)


class LoadView(APIView):
    """
    Expect the upload ID to be provided with the 'id' parameter
    This may be either an upload_id that is stored in the StoredUpload
    table or it may be the path to a file (relative to the fixed upload
    directory specified by the DJANGO_DRF_FILEPOND_FILE_STORE_PATH
    setting parameter).
    """
    permission_classes = _import_permission_classes('GET_LOAD')

    def get(self, request):
        LOG.debug('Filepond API: Load view GET called...')

        if LOAD_RESTORE_PARAM_NAME not in request.GET:
            return Response('A required parameter is missing.',
                            status=status.HTTP_400_BAD_REQUEST)

        upload_id = request.GET[LOAD_RESTORE_PARAM_NAME]

        if (not upload_id) or (upload_id == ''):
            return Response('An invalid ID has been provided.',
                            status=status.HTTP_400_BAD_REQUEST)

        try:
            su = get_stored_upload(upload_id)
        except StoredUpload.DoesNotExist as e:
            LOG.error('StoredUpload with ID [%s] not found: [%s]'
                      % (upload_id, str(e)))
            return Response('Not found', status=status.HTTP_404_NOT_FOUND)

        # su is now the StoredUpload record for the requested file
        try:
            (filename, data_bytes) = get_stored_upload_file_data(su)
        except ConfigurationError as e:
            LOG.error('Error getting file upload: [%s]' % str(e))
            return HttpResponseServerError('The file upload settings are '
                                           'not configured correctly.')
        except FileNotFoundError:
            return HttpResponseNotFound('Error accessing file, not found.')
        except IOError:
            return HttpResponseServerError('Error reading file...')

        ct = _get_content_type(filename)

        response = HttpResponse(data_bytes, content_type=ct)
        response['Content-Disposition'] = ('inline; filename=%s' %
                                           filename)

        return response


class RestoreView(APIView):
    permission_classes = _import_permission_classes('GET_RESTORE')

    # Expect the upload ID to be provided with the 'name' parameter
    def get(self, request):
        LOG.debug('Filepond API: Restore view GET called...')
        if LOAD_RESTORE_PARAM_NAME not in request.GET:
            return Response('A required parameter is missing.',
                            status=status.HTTP_400_BAD_REQUEST)

        upload_id = request.GET[LOAD_RESTORE_PARAM_NAME]

        upload_id_fmt = re.compile('^([%s]){22}$' %
                                   (shortuuid.get_alphabet()))

        if not upload_id_fmt.match(upload_id):
            return Response('An invalid ID has been provided.',
                            status=status.HTTP_400_BAD_REQUEST)

        LOG.debug('Carrying out restore for file ID <%s>' % upload_id)

        try:
            tu = TemporaryUpload.objects.get(upload_id=upload_id)
        except TemporaryUpload.DoesNotExist:
            return Response('Not found', status=status.HTTP_404_NOT_FOUND)

        upload_file_name = tu.upload_name
        try:
            with open(tu.file.path, 'rb') as f:
                data = f.read()
        except IOError as e:
            LOG.error('Error reading requested file: %s' % str(e))
            return Response('Error reading file data...',
                            status=status.HTTP_500_INTERNAL_SERVER_ERROR)

        ct = _get_content_type(upload_file_name)

        response = HttpResponse(data, content_type=ct)
        response['Content-Disposition'] = ('inline; filename=%s' %
                                           upload_file_name)

        return response


class FetchView(APIView):
    permission_classes = _import_permission_classes('GET_FETCH')

    def _process_request(self, request):
        LOG.debug('Filepond API: Fetch view GET called...')
        '''
        Supports retrieving a file on the server side that the user has
        specified by calling addFile on the filepond API and passing a
        URL to a file.
        '''
        # Retrieve the target URL from the request query string target
        # target parameter, pull the file into temp upload storage and
        # return a file object.

        # First check we have a URL and parse to check it's valid
        target_url = request.query_params.get('target', None)
        if not target_url:
            raise ParseError('Required query parameter(s) missing.')

        # Use Django's URL validator to see if we've been given a valid URL
        validator = URLValidator(message=('An invalid URL <%s> has been '
                                          'provided' % (target_url)))
        try:
            validator(target_url)
        except ValidationError as e:
            raise ParseError(str(e))

        # TODO: SHould we check the headers returned when we request the
        # download to see that we're getting a file rather than an HTML page?
        # For now this check is enabled on the basis that we assume target
        # data file will not be HTML. However, there should be a way to turn
        # this off if the client knows that they want to get an HTML file.
        # TODO: *** Looks like this use of head can be removed since with
        # the new approach of streaming content to a BytesIO object, when
        # stream=True, the connection begins by being opened and only
        # fetching the headers. We could do this check then.
        try:
            header = requests.head(target_url, allow_redirects=True)
        except ConnectionError as e:
            msg = ('Unable to access the requested remote file headers: %s'
                   % str(e))
            LOG.error(msg)
            return Response(msg, status=status.HTTP_500_INTERNAL_SERVER_ERROR)

        if header.status_code == 404:
            raise NotFound('The remote file was not found.')

        content_type = header.headers.get('Content-Type', '')

        # If the URL has returned URL content but an HTML file was not
        # requested then assume that the URL has linked to a download page or
        # some sort of error page or similar and raise an error.
        if 'html' in content_type.lower() and '.html' not in target_url:
            LOG.error('The requested data seems to be in HTML format. '
                      'Assuming this is not valid data file.')
            raise ParseError('Provided URL links to HTML content.')

        buf = BytesIO()
        upload_file_name = None
        try:
            with requests.get(target_url,
                              allow_redirects=True, stream=True) as r:
                if 'Content-Disposition' in r.headers:
                    cd = r.headers['Content-Disposition']
                    matches = re.findall('filename=(.+)', cd)
                    if len(matches):
                        upload_file_name = matches[0]
                for chunk in r.iter_content(chunk_size=1048576):
                    buf.write(chunk)
        except ConnectionError as e:
            raise NotFound('Unable to access the requested remote file: %s'
                           % str(e))

        file_id = _get_file_id()
        # If filename wasn't extracted from Content-Disposition header, get
        # from the URL or otherwise set it to the auto-generated file_id
        if not upload_file_name:
            if not target_url.endswith('/'):
                split = target_url.rsplit('/', 1)
                upload_file_name = split[1] if len(split) > 1 else split[0]
            else:
                upload_file_name = file_id

        return (buf, file_id, upload_file_name, content_type)

    def head(self, request):
        LOG.debug('Filepond API: Fetch view HEAD called...')
        result = self._process_request(request)
        if isinstance(result, tuple):
            buf, file_id, upload_file_name, content_type = result
        elif isinstance(result, Response):
            return result
        else:
            raise ValueError('process_request result is of an unexpected type')

        file_size = buf.seek(0, os.SEEK_END)
        buf.seek(0)

        # The addressing of filepond issue #154
        # (https://github.com/pqina/filepond/issues/154) means that fetch
        # can now store a file downloaded from a remote URL and return file
        # metadata in the header if a HEAD request is received. If we get a
        # GET request then the standard approach of proxying the file back
        # to the client is used.
        upload_id = _get_file_id()
        memfile = InMemoryUploadedFile(buf, None, file_id, content_type,
                                       file_size, None)
        tu = TemporaryUpload(upload_id=upload_id, file_id=file_id,
                             file=memfile, upload_name=upload_file_name,
                             upload_type=TemporaryUpload.URL,
                             uploaded_by=_get_user(request))
        tu.save()

        response = Response(status=status.HTTP_200_OK)
        response['Content-Type'] = content_type
        response['Content-Length'] = file_size
        response['X-Content-Transfer-Id'] = upload_id
        response['Content-Disposition'] = ('inline; filename=%s' %
                                           upload_file_name)
        return response

    def get(self, request):
        result = self._process_request(request)
        if isinstance(result, tuple):
            buf, _, upload_file_name, content_type = result
        elif isinstance(result, Response):
            return result
        else:
            raise ValueError('process_request result is of an unexpected type')
        response = HttpResponse(buf.getvalue(), content_type=content_type)
        response['Content-Disposition'] = ('inline; filename=%s' %
                                           upload_file_name)
        return response



# Generated by Django 2.2.8 on 2019-12-17 21:05
from django.conf import settings
from django.db import migrations, models
import django.db.models.deletion


class Migration(migrations.Migration):

    dependencies = [
        migrations.swappable_dependency(settings.AUTH_USER_MODEL),
        ('django_drf_filepond', '0004_tempupload_mods'),
    ]

    operations = [
        migrations.AddField(
            model_name='storedupload',
            name='uploaded_by',
            field=models.ForeignKey(
                blank=True, null=True,
                on_delete=django.db.models.deletion.CASCADE,
                to=settings.AUTH_USER_MODEL),
        ),
        migrations.AddField(
            model_name='temporaryupload',
            name='uploaded_by',
            field=models.ForeignKey(
                blank=True, null=True,
                on_delete=django.db.models.deletion.CASCADE,
                to=settings.AUTH_USER_MODEL),
        ),
    ]



# Generated by Django 3.0.5 on 2021-06-25 16:33

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('django_drf_filepond', '0009_temporaryuploadchunked_offset_utime'),
    ]

    operations = [
        migrations.AlterField(
            model_name='temporaryuploadchunked',
            name='offset',
            field=models.BigIntegerField(default=0),
        ),
        migrations.AlterField(
            model_name='temporaryuploadchunked',
            name='total_size',
            field=models.BigIntegerField(default=0),
        ),
    ]



# Generated by Django 2.2.5 on 2020-04-17 15:38

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('django_drf_filepond', '0008_temporaryuploadchunked'),
    ]

    operations = [
        migrations.AddField(
            model_name='temporaryuploadchunked',
            name='offset',
            field=models.IntegerField(default=0),
        ),
        migrations.AlterField(
            model_name='temporaryuploadchunked',
            name='last_upload_time',
            field=models.DateTimeField(auto_now=True),
        ),
    ]



# A module containing some utility functions used by the views and uploaders
import django_drf_filepond.drf_filepond_settings as local_settings
from django.contrib.auth.models import AnonymousUser
import shortuuid
import six


# Get the user associated with the provided request. If we have an anonymous
# user object then return None
def _get_user(request):
    upload_user = getattr(request, 'user', None)
    if isinstance(upload_user, AnonymousUser):
        upload_user = None
    return upload_user


# Generate a file or upload ID. At present, this is used for generating both
# ID types since they both have the same specification.
def _get_file_id():
    file_id = shortuuid.uuid()
    return six.ensure_text(file_id)


# Get the BASE_DIR variable from local_settings and process it to ensure that
# it can be used in django_drf_filepond across Python 2.7, 3.5 and 3.6+.
# Need to take into account that this may be a regular string or a
# pathlib.Path object. django-drf-filepond expects to work with BASE_DIR as a
# string so return a string regardless of the type of BASE_DIR. To maintain
# suport for Python 2.7, need to handle the case where pathlib.Path doesn't
# exist...
def get_local_settings_base_dir():
    base_dir = local_settings.BASE_DIR
    return _process_base_dir(base_dir)


# Process the provided BASE_DIR variable
def _process_base_dir(base_dir):
    try:
        from pathlib import Path
    except ImportError:
        return base_dir

    if isinstance(base_dir, Path):
        return str(base_dir)
    return base_dir



# -*- coding: utf-8 -*-
# Generated by Django 1.11.14 on 2018-10-24 17:59
from __future__ import unicode_literals

import django.core.files.storage
import django.core.validators
from django.db import migrations, models


class Migration(migrations.Migration):

    initial = True

    dependencies = [
    ]

    operations = [
        migrations.CreateModel(
            name='TemporaryUpload',
            fields=[
                ('file_id',
                 models.CharField(
                     max_length=22, primary_key=True, serialize=False,
                     validators=[
                         django.core.validators.MinLengthValidator(22)])),
                ('file',
                 models.FileField(
                     storage=django.core.files.storage.FileSystemStorage(
                         location=b'/tmp/filepond_uploads'), upload_to=b'')),
                ('upload_name', models.CharField(max_length=512)),
                ('uploaded', models.DateTimeField(auto_now_add=True)),
                ('upload_type',
                 models.CharField(
                     choices=[('F', 'Uploaded file data'),
                              ('U', 'Remote file URL')],
                     max_length=1)),
            ],
        ),
    ]



# Generated by Django 2.2.5 on 2020-04-16 15:21

from django.conf import settings
import django.core.validators
from django.db import migrations, models
import django.db.models.deletion


class Migration(migrations.Migration):

    dependencies = [
        migrations.swappable_dependency(settings.AUTH_USER_MODEL),
        ('django_drf_filepond', '0007_altered_stored_upload_file'),
    ]

    operations = [
        migrations.CreateModel(
            name='TemporaryUploadChunked',
            fields=[
                ('upload_id',
                 models.CharField(
                     max_length=22, primary_key=True, serialize=False,
                     validators=[
                         django.core.validators.MinLengthValidator(22)])),
                ('file_id',
                 models.CharField(
                     max_length=22,
                     validators=[
                         django.core.validators.MinLengthValidator(22)])),
                ('upload_dir',
                 models.CharField(
                     default=models.CharField(
                         max_length=22, primary_key=True, serialize=False,
                         validators=[
                             django.core.validators.MinLengthValidator(22)]),
                     max_length=512)),
                ('last_chunk', models.IntegerField(default=0)),
                ('total_size', models.IntegerField(default=0)),
                ('upload_name', models.CharField(default='', max_length=512)),
                ('upload_complete', models.BooleanField(default=False)),
                ('last_upload_time', models.DateTimeField(auto_now_add=True)),
                ('uploaded_by',
                 models.ForeignKey(
                     blank=True, null=True,
                     on_delete=django.db.models.deletion.CASCADE,
                     to=settings.AUTH_USER_MODEL)),
            ],
        ),
    ]



# Generated by Django 3.0.3 on 2020-02-21 18:59

from django.db import migrations, models
import django_drf_filepond.models


class Migration(migrations.Migration):

    dependencies = [
        ('django_drf_filepond', '0006_permupload_mods'),
    ]

    operations = [
        migrations.AlterField(
            model_name='storedupload',
            name='file',
            field=models.FileField(
                max_length=2048,
                storage=django_drf_filepond.models.DrfFilePondStoredStorage(),
                upload_to=''),
        ),
    ]



# This module contains API functions exported through the top-level
# django_drf_filepond module.
#
# store_upload: used to move an upload from temporary storage to a permanent
#               storage location. If you're using local file storage, this
#               requires that you have the DJANGO_DRF_FILEPOND_FILE_STORE_PATH
#               setting set in your application's settings.py file.
#
import logging
import ntpath
import os
import shutil

import django_drf_filepond.drf_filepond_settings as local_settings
from django.core.exceptions import ImproperlyConfigured
import re
import shortuuid
from django_drf_filepond.models import TemporaryUpload, StoredUpload
from django_drf_filepond.storage_utils import _get_storage_backend
from django_drf_filepond.exceptions import ConfigurationError

# TODO: Need to refactor this into a class and put the initialisation of
# the storage backend into the init.
storage_backend_initialised = False
storage_backend = None

LOG = logging.getLogger(__name__)

# There's no built in FileNotFoundError, FileExistsError in Python 2
try:
    FileNotFoundError
except NameError:
    FileNotFoundError = IOError

try:
    FileExistsError
except NameError:
    FileExistsError = OSError


def _init_storage_backend():
    global storage_backend_initialised
    global storage_backend

    storage_module_name = getattr(local_settings, 'STORAGES_BACKEND', None)
    LOG.debug('Initialising storage backend with storage module name [%s]'
              % storage_module_name)
    storage_backend = _get_storage_backend(storage_module_name)
    storage_backend_initialised = True


# Store the temporary upload represented by upload_id to the specified
# destination_file_path under the defined file store location as specified by
# the DJANGO_DRF_FILEPOND_FILE_STORE_PATH configuration setting (for local
# storage). Files stored using this approach can subsequently be retrieved
# using the load method defined in by the filepond server spec by using
# either the 22-char upload_id or the value provided to the
# destination_file_path parameter as a query string parameter using the
# "id" key.
def store_upload(upload_id, destination_file_path):
    """
    Store the temporary upload with the specified upload ID to the
    destination_file_path. destination_file_path should be a directory only
    and not include the target name of the file.

    If destination_file_name is not provided, the file
    is stored using the name it was originally uploaded with. If
    destination_file_name is provided, this is the name used to store the
    file. i.e. the file will be stored at
        destination_file_path + destination_file_name
    """
    # TODO: If the storage backend is not initialised, init now - this will
    # be removed when this module is refactored into a class.
    if not storage_backend_initialised:
        _init_storage_backend()

    # If there's no storage backend set then we're using local file storage
    # and FILE_STORE_PATH must be set.
    if not storage_backend:
        if ((not hasattr(local_settings, 'FILE_STORE_PATH')) or
                (not local_settings.FILE_STORE_PATH)):
            raise ImproperlyConfigured('A required setting is missing in your '
                                       'application configuration.')

    id_fmt = re.compile('^([%s]){22}$' % (shortuuid.get_alphabet()))
    if not id_fmt.match(upload_id):
        LOG.error('The provided upload ID <%s> is of an invalid format.'
                  % upload_id)
        raise ValueError('The provided upload ID is of an invalid format.')

    if not destination_file_path or destination_file_path == '':
        raise ValueError('No destination file path provided.')

    try:
        tu = TemporaryUpload.objects.get(upload_id=upload_id)
    except TemporaryUpload.DoesNotExist:
        raise ValueError('Record for the specified upload_id doesn\'t exist')

    # Before this was updated, passing a path ending in os.sep, i.e. a
    # directory name, would ensure that the file was stored in the specified
    # directory using the name that the file had when it was originally
    # uploaded.  To ensure consistency with this previous approach to
    # handling files, we have to ensure here that if the original path ends
    # in os.sep and we're using local storage, this is maintained when it
    # is passed to _store_upload_local
    destination_name = ntpath.basename(destination_file_path)
    destination_path = ntpath.dirname(destination_file_path)

    if ((not storage_backend) and (destination_name == '') and
            (destination_file_path.endswith(os.sep))):
        # In some cases we'll enter this block but destination path will
        # already end in a '/' so check before updating
        if not destination_path.endswith('/'):
            destination_path += os.sep

    if storage_backend:
        return _store_upload_remote(destination_path, destination_name, tu)
    else:
        return _store_upload_local(destination_path, destination_name, tu)


def _store_upload_local(destination_file_path, destination_file_name,
                        temp_upload):
    file_path_base = local_settings.FILE_STORE_PATH

    # If called via store_upload, this has already been checked but in
    # case this is called directly, double check that the store path is set
    if not file_path_base or file_path_base == '':
        raise ValueError('The FILE_STORE_PATH is not set to a directory.')

    # Is this necessary? Checking on every file storage in case the directory
    # was removed but not sure that this is really necessary.
    if((not os.path.exists(file_path_base)) or
            (not os.path.isdir(file_path_base))):
        raise FileNotFoundError(
            'The local output directory [%s] defined by FILE_STORE_PATH is '
            'missing.' % file_path_base)

    if destination_file_path.startswith(os.sep):
        destination_file_path = destination_file_path[1:]

    target_dir = os.path.join(file_path_base, destination_file_path)
    target_filename = destination_file_name
    # If not filename provided, assume a directory was provided, get the
    # file name from temp_upload and use this
    if not target_filename:
        target_filename = temp_upload.upload_name
    destination_file_path = os.path.join(destination_file_path,
                                         target_filename)

    # Check we're not about to overwrite anything
    target_file_path = os.path.join(target_dir, target_filename)
    if os.path.exists(target_file_path):
        LOG.error('File with specified name and path <%s> already exists'
                  % target_file_path)
        raise FileExistsError('The specified temporary file cannot be stored'
                              ' to the specified location - file exists.')

    su = StoredUpload(upload_id=temp_upload.upload_id,
                      file=destination_file_path,
                      uploaded=temp_upload.uploaded,
                      uploaded_by=temp_upload.uploaded_by)

    try:
        if not os.path.exists(target_dir):
            os.makedirs(target_dir)
        shutil.copy2(temp_upload.get_file_path(), target_file_path)
        su.save()
        temp_upload.delete()
    except IOError as e:
        LOG.error('Error moving temporary file to permanent storage location')
        raise e

    return su


def _store_upload_remote(destination_file_path, destination_file_name,
                         temp_upload):
    # Use the storage backend to write the file to the storage backend
    target_filename = destination_file_name
    if not target_filename:
        target_filename = temp_upload.upload_name

    su = None
    destination_file = os.path.join(destination_file_path, target_filename)
    try:
        storage_backend.save(destination_file, temp_upload.file)
        su = StoredUpload(upload_id=temp_upload.upload_id,
                          file=destination_file,
                          uploaded=temp_upload.uploaded,
                          uploaded_by=temp_upload.uploaded_by)
        su.save()
        temp_upload.delete()
    except Exception as e:
        errorMsg = ('Error storing temporary upload to remote storage: [%s]'
                    % str(e))
        LOG.error(errorMsg)
        raise e

    return su


def get_stored_upload(upload_id):
    """
    Get an upload that has previously been stored using the store_upload
    function.

    upload_id: This function takes a 22-character unique ID assigned to the
    original upload of the requested file.
    """
    # If the parameter matched the upload ID format, we assume that it
    # must be an upload ID and proceed accordingly. If the lookup of the
    # record fails, then we have another go assuming a filename was
    # instead provided.

    # NOTE: The API doesn't officially provide support for requesting stored
    # uploads by filename. This is retained here for backward compatibility
    # but it is DEPRECATED and will be removed in a future release.
    param_filename = False

    upload_id_fmt = re.compile('^([%s]){22}$'
                               % (shortuuid.get_alphabet()))

    if not upload_id_fmt.match(upload_id):
        param_filename = True
        LOG.debug('The provided string doesn\'t seem to be an '
                  'upload ID. Assuming it is a filename/path.')

    if not param_filename:
        try:
            su = StoredUpload.objects.get(upload_id=upload_id)
        except StoredUpload.DoesNotExist:
            LOG.debug('A StoredUpload with the provided ID doesn\'t '
                      'exist. Assuming this could be a filename.')
            param_filename = True

    if param_filename:
        # Try and lookup a StoredUpload record with the specified id
        # as the file path
        try:
            su = StoredUpload.objects.get(file=upload_id)
        except StoredUpload.DoesNotExist as e:
            LOG.debug('A StoredUpload with the provided file path '
                      'doesn\'t exist. Re-raising error')
            raise e

    return su


def get_stored_upload_file_data(stored_upload):
    """
    Given a StoredUpload object, this function gets and returns the data of
    the file associated with the StoredUpload instance.

    This function provides an abstraction over the storage backend, accessing
    the file data regardless of whether the file is stored on the local
    filesystem or on some remote storage service, e.g. Amazon S3. Supported
    storage backends are those supported by the django-storages library.

    Returns a tuple (filename, data_bytes_io).
        filename is a string containing the name of the stored file
        data_bytes_io is a file-like BytesIO object containing the file data
    """
    # TODO: If the storage backend is not initialised, init now - this
    # will be removed when this module is refactored into a class.
    if not storage_backend_initialised:
        _init_storage_backend()
    if storage_backend:
        LOG.debug('get_stored_upload_file_data: Using a remote storage '
                  'service: [%s]' % (type(storage_backend).__name__))

        file_path_base = ''
    else:
        LOG.debug('get_stored_upload_file_data: Using local storage backend.')
        if ((not hasattr(local_settings, 'FILE_STORE_PATH')) or
                (not local_settings.FILE_STORE_PATH) or
                (not os.path.exists(local_settings.FILE_STORE_PATH)) or
                (not os.path.isdir(local_settings.FILE_STORE_PATH))):
            raise ConfigurationError('The file upload settings are not '
                                     'configured correctly.')

        file_path_base = local_settings.FILE_STORE_PATH
        #  This code is redundant, this case will be picked up by the
        #  not local_settings.FILE_STORE_PATH in the above statement.
        #   if not file_path_base:
        #       file_path_base = ''

    # See if the stored file with the path specified in su exists
    # in the file store location
    file_path = os.path.join(file_path_base, stored_upload.file.name)
    if storage_backend:
        if not storage_backend.exists(file_path):
            LOG.error('File [%s] for upload_id [%s] not found on remote '
                      'file store' % (file_path, stored_upload.upload_id))
            raise FileNotFoundError(
                'File [%s] for upload_id [%s] not found on remote file '
                'store.' % (file_path, stored_upload.upload_id))
        file_data = stored_upload.file.read()
    else:
        if ((not os.path.exists(file_path)) or
                (not os.path.isfile(file_path))):
            LOG.error('File [%s] for upload_id [%s] not found on local disk'
                      % (file_path, stored_upload.upload_id))
            raise FileNotFoundError('File [%s] not found on local disk'
                                    % file_path)

        # We now know that the file exists locally and is not a directory
        file_data = stored_upload.file.read()

    filename = os.path.basename(stored_upload.file.name)
    return (filename, file_data)


def delete_stored_upload(upload_id, delete_file=False):
    """
    Delete the specified stored upload AND IF delete_file=True ALSO
    PERMANENTLY DELETE THE FILE ASSOCIATED WITH THE UPLOAD.

    It is necessary to explicitly set delete_file=True to ensure that it
    is made explicit that the stored file associated with the upload will be
    permanently deleted.
    """
    try:
        su = get_stored_upload(upload_id)
    except StoredUpload.DoesNotExist as e:
        LOG.error('No stored upload found with the specified ID [%s].'
                  % (upload_id))
        raise e

    # Need to retain upload ID here since this is used in error messages later
    upload_id = su.upload_id

    su.delete()

    if not delete_file:
        return True

    # If we got the stored file record and delete_file is True, make sure
    # that the storage backend is set up and we have access to it.
    # TODO: If the storage backend is not initialised, init now - this
    # will be removed when this module is refactored into a class.
    if not storage_backend_initialised:
        _init_storage_backend()

    if storage_backend:
        LOG.debug('delete_stored_upload: Using a remote storage '
                  'service: [%s]' % (type(storage_backend).__name__))
        file_path_base = ''
    else:
        LOG.debug('delete_stored_upload: Using local storage backend.')
        if ((not hasattr(local_settings, 'FILE_STORE_PATH')) or
                (not local_settings.FILE_STORE_PATH) or
                (not os.path.exists(local_settings.FILE_STORE_PATH)) or
                (not os.path.isdir(local_settings.FILE_STORE_PATH))):
            raise ConfigurationError('The file upload settings are not '
                                     'configured correctly.')

        file_path_base = local_settings.FILE_STORE_PATH

    file_path = os.path.join(file_path_base, su.file.name)
    if storage_backend:
        if not storage_backend.exists(file_path):
            LOG.error('Stored upload file [%s] with upload_id [%s] is not '
                      'found on remote file store' % (file_path, upload_id))
            raise FileNotFoundError(
                'File [%s] for stored upload with id [%s] not found on remote'
                ' file store.' % (file_path, upload_id))
        storage_backend.delete(file_path)
    # Else delete local file
    else:
        if ((not os.path.exists(file_path)) or
                (not os.path.isfile(file_path))):
            LOG.error('File [%s] for stored upload [%s] not found on '
                      'local disk' % (file_path, upload_id))
            raise FileNotFoundError('File [%s] to delete was not found on '
                                    'the local disk' % file_path)

        # We now know that the file exists locally and is not a directory
        try:
            os.remove(file_path)
        except OSError as e:
            LOG.error('Error removing requested file: %s' % str(e))
            raise e

        # TODO: Need to look at how best to delete directories that may have
        # been created to store the file. For now, we just delete the file.

    return True



'''
Created on 24 Oct 2018

@author: jcohen02

A renderers module to host a PlainTextRenderer that will render
plain text responses sending back the raw text to the client.
'''
from collections import OrderedDict
import json
import logging

from rest_framework.renderers import BaseRenderer


LOG = logging.getLogger(__name__)


# This plaintext renderer is taken from the example in the
# django rest framework docs since this provides exactly what we
# require but doesn't seem to be included in the core DRF API. See:
# https://www.django-rest-framework.org/api-guide/renderers/#custom-renderers
# This renderer avoids the issue with the standard JSONRenderer that
# results in raw text responses being wrapped in quotes.
class PlainTextRenderer(BaseRenderer):
    '''
    Plain text renderer.
    '''
    media_type = 'text/plain'
    format = 'txt'

    def render(self, data, media_type=None, renderer_context=None):
        '''
        Encode the raw data - default charset is UTF-8.
        '''
        LOG.debug('Data is <%s>' % data)

        if data:
            if type(data) in [dict, OrderedDict]:
                return json.dumps(data)
            else:
                return data.encode(self.charset)
        return data



from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('django_drf_filepond', '0005_add_uploaded_by'),
    ]

    operations = [
        migrations.AlterField(
            model_name='storedupload',
            name='file_path',
            field=models.FileField(max_length=2048, upload_to=''),
        ),
        migrations.RenameField(
            model_name='storedupload',
            old_name='file_path',
            new_name='file'
        )
    ]



# Generated by Django 2.1.3 on 2019-08-03 17:03
import django.core.validators
from django.db import migrations, models
import django_drf_filepond.models as drfmodels


class Migration(migrations.Migration):

    dependencies = [
        ('django_drf_filepond', '0003_add_storedupload'),
    ]

    operations = [
        migrations.AlterField(
            model_name='temporaryupload',
            name='file',
            field=models.FileField(
                storage=drfmodels.FilePondUploadSystemStorage(),
                upload_to=drfmodels.get_upload_path),
        ),
        migrations.AlterField(
            model_name='temporaryupload',
            name='file_id',
            field=models.CharField(
                max_length=22,
                validators=[django.core.validators.MinLengthValidator(22)]),
        ),
    ]



'''
Setup some local application settings with local defaults.
This approach is based on the one shown in this blog post:

http://blog.muhuk.com/2010/01/26/developing-reusable-django-apps-app-settings.html#.W_Rkh-mYQ6U

******** Settings in this file are not currently being used for core library
******** use due to an issue with dynamically setting up the storage
******** location in the FileSystemStorage class. For now, you must set
******** DJANGO_DRF_FILEPOND_UPLOAD_TMP in your top level app settings.
******** THIS IS CURRENTLY USED ONLY FOR INTEGRATION TESTS
'''
import django_drf_filepond
import os
from django.conf import settings

_app_prefix = 'DJANGO_DRF_FILEPOND_'

# BASE_DIR is assumed to be present in the core project settings. However,
# in case it isn't, we check here and set a local BASE_DIR to the
# installed app base directory to use as an alternative.
BASE_DIR = os.path.dirname(django_drf_filepond.__file__)
if hasattr(settings, 'BASE_DIR'):
    # If BASE_DIR is set in the main settings, get it and process it to
    # handle py3.5 where pathlib exists but os.path.join can't accept a
    # pathlib object (ensure we always pass a string to os.path.join)
    BASE_DIR = settings.BASE_DIR
    try:
        from pathlib import Path
        if isinstance(BASE_DIR, Path):
            BASE_DIR = str(BASE_DIR)
    except ImportError:
        pass

# The location where uploaded files are temporarily stored. At present,
# this must be a subdirectory of settings.BASE_DIR
UPLOAD_TMP = getattr(settings, _app_prefix+'UPLOAD_TMP',
                     os.path.join(BASE_DIR, 'filepond_uploads'))

# Setting to control whether the temporary directories created for file
# uploads are removed when an uploaded file is deleted
DELETE_UPLOAD_TMP_DIRS = getattr(settings,
                                 _app_prefix+'DELETE_UPLOAD_TMP_DIRS', True)

# Specifies the django-storages backend to be used. See the list at:
# https://django-storages.readthedocs.io
# If this is not set, then the default local filesystem backend is used.
# If you set this parameter, you also need to set the relevant parameters
# for your chosen backend as described in the django-storages documentation.
STORAGES_BACKEND = getattr(settings, _app_prefix+'STORAGES_BACKEND', None)

# The file storage location used by the top-level application. This needs to
# be set if the load endpoint is going to be used to access files that have
# been permanently stored after being uploaded as TemporaryUpload objects.
# If you're using django-storages, this path is the base path to be used
# on the backend storage.
# If STORAGES_BACKEND is provided, this MUST be set
# If you're not using a STORAGES_BACKEND and this is NOT set, you can't use
# django-drf-filepond's file management and the api.store_upload function
# will not be usable - you will need to manage file storage in your code.
FILE_STORE_PATH = getattr(settings, _app_prefix+'FILE_STORE_PATH', None)

# If you want to use an external directory (a directory outside of your
# project directory) to store temporary uploads, this setting needs to be
# set to true. By default it is False to prevent uploads being stored
# elsewhere.
ALLOW_EXTERNAL_UPLOAD_DIR = getattr(settings,
                                    _app_prefix+'ALLOW_EXTERNAL_UPLOAD_DIR',
                                    False)
# Optional permissions settings for each endpoint.
PERMISSION_CLASSES = getattr(settings, _app_prefix+'PERMISSION_CLASSES', {})



import importlib
import logging


LOG = logging.getLogger(__name__)


def _get_storage_backend(fq_classname):
    """
    Load the specified django-storages storage backend class. This is called
    regardless of whether a beckend is specified so if fq_classname is not
    set, we just return None.

    fq_classname is a string specifying the fully-qualified class name of
    the django-storages backend to use, e.g.
        'storages.backends.sftpstorage.SFTPStorage'
    """
    LOG.debug('Running _get_storage_backend with fq_classname [%s]'
              % fq_classname)

    if not fq_classname:
        return None

    (modname, clname) = fq_classname.rsplit('.', 1)
    # A test import of the backend storage class should have been undertaken
    # at app startup in django_drf_filepond.apps.ready so any failure
    # importing the backend should have been picked up then.
    mod = importlib.import_module(modname)
    storage_backend = getattr(mod, clname)()
    LOG.info('Storage backend instance [%s] created...' % fq_classname)

    return storage_backend



class ConfigurationError(Exception):
    '''
    Raised when a problem occurs with the configuration of the library.
    '''
    pass


class APIError(Exception):
    '''
    Raised when a problem occurs in the API functions.
    '''
    pass



from django.apps import AppConfig

import importlib
import os
import logging
import django_drf_filepond.drf_filepond_settings as local_settings
from django.core.exceptions import ImproperlyConfigured

LOG = logging.getLogger(__name__)


class DjangoDrfFilepondConfig(AppConfig):
    name = 'django_drf_filepond'
    verbose_name = 'FilePond Server-side API'

    def ready(self):
        # Get BASE_DIR and process to ensure it works across platforms
        # Handle py3.5 where pathlib exists but os.path.join can't accept a
        # pathlib object (ensure we always pass a string to os.path.join)
        BASE_DIR = local_settings.BASE_DIR
        try:
            from pathlib import Path
            if isinstance(BASE_DIR, Path):
                BASE_DIR = str(BASE_DIR)
        except ImportError:
            pass

        upload_tmp = getattr(
            local_settings, 'UPLOAD_TMP',
            os.path.join(BASE_DIR, 'filepond_uploads'))

        LOG.debug('Upload temp directory from top level settings: <%s>'
                  % (upload_tmp))

        # Check if the temporary file directory is available and if not
        # create it.
        if not os.path.exists(local_settings.UPLOAD_TMP):
            LOG.warning('Filepond app init: Creating temporary file '
                        'upload directory <%s>' % local_settings.UPLOAD_TMP)
            os.makedirs(local_settings.UPLOAD_TMP, mode=0o700)
        else:
            LOG.debug('Filepond app init: Temporary file upload '
                      'directory already exists')

        # See if we're using a local file store or django-storages
        # If the latter, we create an instance of the storage class
        # to make sure that it's available and dependencies are installed
        storage_class = getattr(local_settings, 'STORAGES_BACKEND', None)
        if storage_class:
            LOG.info('Using django-storages with backend [%s]'
                     % storage_class)
            (modname, clname) = storage_class.rsplit('.', 1)
            # Either the module import or the getattr to instantiate the
            # class will throw an exception if there's a problem
            # creating storage backend instance due to missing configuration
            # or dependencies.
            mod = importlib.import_module(modname)
            getattr(mod, clname)()
            LOG.info('Storage backend [%s] is available...' % storage_class)
        else:
            LOG.info('App init: no django-storages backend configured, '
                     'using default (local) storage backend if set, '
                     'otherwise you need to manage file storage '
                     'independently of this app.')

        file_store = getattr(local_settings, 'FILE_STORE_PATH', None)
        if file_store:
            if not os.path.exists(file_store):
                LOG.warning('Filepond app init: Creating file store '
                            'directory <%s>...' % file_store)
                os.makedirs(file_store, mode=0o700)
            else:
                LOG.debug('Filepond app init: File store path already '
                          'exists')
        else:
            if not storage_class:
                LOG.error('You are using local file storage so you must set '
                          'the base file storage path using %sFILE_STORE_PATH'
                          % local_settings._app_prefix)
                raise ImproperlyConfigured(
                    'You are using local file storage so you must set the '
                    'base file storage path using %sFILE_STORE_PATH'
                    % local_settings._app_prefix)



# Generated by Django 2.1.3 on 2018-12-04 14:50
import django.core.validators
from django.db import migrations, models
import shortuuid


class Migration(migrations.Migration):

    dependencies = [
        ('django_drf_filepond', '0001_initial'),
    ]

    operations = [
        migrations.RenameField(
            model_name='temporaryupload',
            old_name='file_id',
            new_name='upload_id',
        ),
        migrations.AddField(
            model_name='temporaryupload',
            name='file_id',
            field=models.CharField(
                default=shortuuid.uuid(), max_length=22, serialize=False,
                validators=[django.core.validators.MinLengthValidator(22)]),
            preserve_default=False,
        ),
    ]



"""FilePond server-side URL configuration

Based on the server-side configuration details
provided at:
https://pqina.nl/filepond/docs/patterns/api/server/#configuration
"""
import six
if six.PY2:
    from django.conf.urls import url
else:
    from django.urls import re_path, path
from django_drf_filepond.views import ProcessView, RevertView, LoadView,\
     RestoreView, FetchView, PatchView

#############################################################################
# PYTHON 2 SUPPORT
# Retaining Python 2 compatibility for now. However, with the switch to
# Django 4.0, django.conf.urls.url has been removed and we need to switch
# to paths - this Python 2 specific block will ultimately be removed when
# Python 2.7 support is dropped along with support for django 1.x.
#
# django-drf-filepond SUPPORT FOR PYTHON 2.7/DJANGO 1.11 WILL BE DEPRECATED
# IN RELEASE 0.5.0. THIS WILL BE THE LAST RELEASE TO SUPPORT PYTHON 2.7.
#############################################################################
if six.PY2:
    urlpatterns = [
        url(r'^process/$', ProcessView.as_view(), name='process'),
        url(r'^patch/(?P<chunk_id>[0-9a-zA-Z]{22})$', PatchView.as_view(),
            name='patch'),
        url(r'^revert/$', RevertView.as_view(), name='revert'),
        url(r'^load/$', LoadView.as_view(), name='load'),
        url(r'^restore/$', RestoreView.as_view(), name='restore'),
        url(r'^fetch/$', FetchView.as_view(), name='fetch')
    ]
else:
    urlpatterns = [
        path('process/', ProcessView.as_view(), name='process'),
        re_path(r'^patch/(?P<chunk_id>[0-9a-zA-Z]{22})$', PatchView.as_view(),
                name='patch'),
        path('revert/', RevertView.as_view(), name='revert'),
        path('load/', LoadView.as_view(), name='load'),
        path('restore/', RestoreView.as_view(), name='restore'),
        path('fetch/', FetchView.as_view(), name='fetch')
    ]



from setuptools import setup

with open("README.md", "r") as readme:
    long_description = readme.read()

setup(
    name="django-drf-filepond",
    version="0.4.1",
    description="Filepond server app for Django REST Framework",
    long_description=long_description,
    long_description_content_type='text/markdown',
    author="Jeremy Cohen",
    author_email="jeremy.cohen@imperial.ac.uk",
    url="https://github.com/ImperialCollegeLondon/django-drf-filepond",
    download_url=(
        "https://github.com/ImperialCollegeLondon/django-drf-filepond.git"),
    license="BSD 3-Clause",
    packages=[
        "django_drf_filepond",
        "django_drf_filepond.migrations",
    ],
    include_package_data=True,
    install_requires=[
        "Django>=1.11.29,<2.0.0;python_version=='2.7'",
        "Django>=2.2.25,<3.0.0;python_version=='3.5'",
        "Django>=2.2.25;python_version>='3.6'",
        "djangorestframework==3.9.4;python_version=='2.7'",
        "djangorestframework>=3.11.2;python_version>='3.5'",
        "shortuuid==0.5.0;python_version=='2.7'",
        "shortuuid>=0.5.0;python_version>='3.5'",
        "requests>=2.20.1",
        "django-storages==1.9.1;python_version=='2.7'",
        "django-storages>=1.9.1;python_version>='3.5'",
        "six>=1.14.0"
    ],
    tests_require=[
        "nose",
        "coverage==5.5;python_version<='3.5'",
        "coverage;python_version>'3.5'",
        "wheel;python_version=='3.5'",
        "httpretty==1.0.3;python_version>='3.5'",
        "httpretty==0.9.7;python_version=='2.7'",
        "mock>=3.0.0,<=3.0.5;python_version<='3.5'",
        "mock>=3.0.5;python_version>'3.5'",
        "paramiko"
    ],
    zip_safe=False,
    test_suite="tests.runner.start",
    classifiers=[
        "Operating System :: OS Independent",
        "Development Status :: 4 - Beta",
        "Environment :: Web Environment",
        "Framework :: Django",
        "Intended Audience :: Developers",
        "License :: OSI Approved :: BSD License",
        "Programming Language :: Python :: 2.7",
        "Programming Language :: Python :: 3.5",
        "Programming Language :: Python :: 3.6",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
        "Topic :: Software Development :: Libraries :: Python Modules",
    ]
)



# Generated by Django 2.1.3 on 2019-02-13 13:00
import django.core.validators
from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('django_drf_filepond', '0002_add_upload_dir'),
    ]

    operations = [
        migrations.CreateModel(
            name='StoredUpload',
            fields=[
                ('upload_id',
                 models.CharField(
                     max_length=22, primary_key=True, serialize=False,
                     validators=[
                         django.core.validators.MinLengthValidator(22)])),
                ('file_path', models.CharField(max_length=2048)),
                ('uploaded', models.DateTimeField()),
                ('stored', models.DateTimeField(auto_now_add=True)),
            ],
        ),
    ]



'''
Created on 24 Oct 2018

@author: jcohen02

A parsers module to host a PlainTextParser that will parse
incoming plain/text requests from filepond
'''
from rest_framework.parsers import BaseParser


# This plaintext parser is taken from the example in the
# django rest framework docs since this provides exactly what we
# require but doesn't seem to be included in the core DRF API.
# See: https://www.django-rest-framework.org/api-guide/parsers/#example
# This will make the data from the body of the request available
# in request.data
class PlainTextParser(BaseParser):
    """
    Plain text parser.
    """
    media_type = 'text/plain'

    def parse(self, stream, media_type=None, parser_context=None):
        """
        Simply return a string representing the body of the request.
        """
        return stream.read()


# The chunk parser is used to parse uploaded file chunks for the chunked
# upload support. A chunk upload request has a content type of
# application/offset+octet-stream. For now we simply get the raw request data
# and return it.
# TODO: This could alse extract metadata from the request, such as chunk
#       length, name and offset and return an object containing the data and
#       the metadata. For now the metadata is extracted and checked prior to
#       accessing the uploaded data.
class UploadChunkParser(BaseParser):
    """
    Upload chunk parser for handling uploaded partial file chunks
    """
    media_type = 'application/offset+octet-stream'

    def parse(self, stream, media_type=None, parser_context=None):
        """
        For now just return the body which contains the uploaded file data
        """
        return stream.read()
